# -*- coding: utf-8 -*-
#
# ---------------------------------------
#   程序：net.py
#   版本：0.2
#   作者：lds
#   日期：2019-01-31
#   语言：Python 3.X
#   说明：网络相关的函数集合
# ---------------------------------------

import os
import uuid
import socket
import struct

BROADCAST_IP = '255.255.255.255'
DEFAULT_PORT = 9


def get_host_ip():
    """
    通过 UDP 获取本机 IP
    print(get_host_ip())
    :return:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip


def get_mac_address():
    """
    获取本机 MAC 地址
    """
    mac = uuid.UUID(int=uuid.getnode()).hex[-12:]
    return ":".join([mac[e:e + 2] for e in range(0, 11, 2)])  # .upper()


def get_arp_list():
    """
    获取局域网 ARP 表

    返回：[[id地址, mac], ]
    """
    host = get_host_ip()
    os.system('arp -a > arp_a.txt')
    arp_list = []
    with open('temp.txt') as fp:
        for line in fp:
            line = line.split()[:2]
            if line and line[0].startswith(host[:4]) and (not line[0].endswith('255')):
                arp_list.append(line)
                # print(':'.join(line))
    return arp_list


# https://github.com/remcohaszing/pywakeonlan
def create_magic_packet(macaddress):
    """
    创建一个魔术包。

    Magic Packet（魔法数据包）是一种数据包，可以与用于唤醒 LAN 协议一起使用来唤醒计算机。 数据包由作为参数给出的 mac 地址构成。

    需要主板的支持，操作系统要求：
    1. 网络属性 》 配置 》 Microsoft 网络客户端 》 属性 》 高级 》 启用 唤醒魔包
    2. 网络属性 》 配置 》 Microsoft 网络客户端 》 属性 》 电源管理 》 取消 允许计算机关闭此设备电源以节约电源

    备注：
    Magic Packet(魔法数据包)。一般通过 UDP 协议进行广播。端口一般是 7 或者 9。
    魔法数据包总是以"FF FF FF FF FF FF" 连续 6 个 "FF"，后面是 MAC 地址信息，一旦网卡侦测到数据包内容，就会唤醒目标计算机。

    # 本地关机命令
    import os
    os.system('shutdown -s -t 00')

    Args:
        macaddress (str): 解析为魔术包的 mac 地址。

    """
    if len(macaddress) == 12:
        pass
    elif len(macaddress) == 17:
        sep = macaddress[2]
        macaddress = macaddress.replace(sep, '')
    else:
        raise ValueError('MAC 地址格式不正确')

    # 填充同步流
    data = b'FFFFFFFFFFFF' + (macaddress * 16).encode()
    send_data = b''

    # 拆分包中的十六进制值
    for i in range(0, len(data), 2):
        send_data += struct.pack(b'B', int(data[i: i + 2], 16))
    return send_data


def send_magic_packet(*macs, **kwargs):
    """
    唤醒局域网中的计算机。

    必须在主机设备上启用 LAN 唤醒。

    Args:
        macs (str): 唤醒机器的一个或多个 mac 地址。

    Keyword Args:
        ip_address (str): 发送魔术包的主机的IP地址 (default "255.255.255.255")
        port (int): 发送魔术包的主机端口 (default 9)


    例子：
    from ilds.net import send_magic_packet

    # 通过其 mac 地址唤醒单台计算机
    send_magic_packet('ff.ff.ff.ff.ff.ff')

    # 通过他们的mac地址唤醒多台计算机。
    send_magic_packet('ff.ff.ff.ff.ff.ff', '00-00-00-00-00-00', 'FFFFFFFFFFFF')

    # 可以指定外部主机。请注意，该主机上的端口转发是必需的。默认IP地址为 255.255.255.255，默认端口为9。
    send_magic_packet('ff.ff.ff.ff.ff.ff', ip_address='example.com', port=1337)
    """
    packets = []
    ip = kwargs.pop('ip_address', BROADCAST_IP)
    port = kwargs.pop('port', DEFAULT_PORT)
    for k in kwargs:
        raise TypeError('send_magic_packet() got an unexpected keyword '
                        'argument {!r}'.format(k))

    for mac in macs:
        packet = create_magic_packet(mac)
        packets.append(packet)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.connect((ip, port))
    for packet in packets:
        sock.send(packet)
    sock.close()


def doc():
    """
    打印模块说明文档
    """
    doc_text = """"""
    doc_text += '\n'
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_host_ip)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_mac_address)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_arp_list)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=send_magic_packet)
    print(doc_text)


if __name__ == '__main__':
    # 记录运行时间 --------------------------------------------------
    from time import time, sleep

    start_time = t1 = time()

    doc()

    print('运行时间 %.2f 秒' % (time() - start_time))



# -*- coding: utf-8 -*-
#
# ---------------------------------------
#   程序：import_export.py
#   版本：0.1
#   作者：lds
#   日期：2019-05-05
#   语言：Python 3.X
#   说明：django 导入和导出
# ---------------------------------------

import csv

# 获取模型实例的字典
from django.forms.models import model_to_dict

from ilds.django.model import get_model_verbose_names
from ilds.django.model import get_model_verbose_name_dict as get_model_field
from ilds.django.model import get_fieldfile_path as get_model_fieldfile_path


class ModelData():
    """
    根据 django 模型转换导入数据
    """

    # 获取字段类型，处理一部分类型
    WIDGETS_MAP = {
        # 'ManyToManyField': 'get_m2m_widget',
        # 'OneToOneField': 'get_fk_widget',
        # 'ForeignKey': 'get_fk_widget',
        # 'DecimalField': widgets.DecimalWidget,
        # 'DateTimeField': widgets.DateTimeWidget,
        # 'DateField': widgets.DateWidget,
        # 'TimeField': widgets.TimeWidget,
        # 'DurationField': widgets.DurationWidget,
        # 'FloatField': float,
        # 'IntegerField': widgets.IntegerWidget,
        # 'PositiveIntegerField': widgets.IntegerWidget,
        # 'BigIntegerField': widgets.IntegerWidget,
        # 'PositiveSmallIntegerField': widgets.IntegerWidget,
        # 'SmallIntegerField': widgets.IntegerWidget,
        # 'AutoField': widgets.IntegerWidget,
        # 'NullBooleanField': widgets.BooleanWidget,
        # 'BooleanField': widgets.BooleanWidget,
    }

    def __init__(self, model, row, import_names=None, exclude=None):

        # 需要转换的 id 然后批量转换
        # 如果有需要 再添加

        # print(filed)

        if exclude is None:
            exclude = []

        self.model = model
        self.import_fields = []
        self.import_index = []
        self.zhuanhuan = {}

        names = {}
        verbose_names = {}
        if import_names is None:
            import_names = {}

        for f in model._meta.fields:
            # print('处理不同字段内容', isinstance(f, models.BooleanField))
            # print('获取模型的字段名', f.get_internal_type())
            internal_type = f.get_internal_type() if callable(getattr(f, "get_internal_type", None)) else ""
            # print(internal_type, f.verbose_name, f.name)
            names[f.name] = {'name': f.name, 'internal_type': internal_type}
            verbose_names[f.verbose_name] = {'name': f.name, 'internal_type': internal_type}

        for i, field in enumerate(row):
            # 排除的内容
            if field in exclude:
                continue

            # 重命名的内容
            if field in import_names:
                field = import_names[field]

            # 字段名导入
            if field in names:
                self.import_fields.append(names[field]['name'])
                self.import_index.append(i)
                if names[field]['internal_type'] in self.WIDGETS_MAP:
                    self.zhuanhuan[names[field]['name']] = self.WIDGETS_MAP[names[field]['internal_type']]
            # 按名字导入
            elif field in verbose_names:
                self.import_fields.append(verbose_names[field]['name'])
                self.import_index.append(i)
                if verbose_names[field]['internal_type'] in self.WIDGETS_MAP:
                    self.zhuanhuan[verbose_names[field]['name']] = self.WIDGETS_MAP[
                        verbose_names[field]['internal_type']]
            # 没有导入的内容
            else:
                # pass
                # print('导入失败：-------------------表头 %s %s 应该为 %s 。-------------------' % (i + 1, field, now_0[i]))
                raise RuntimeError('导入失败：row[%s]：%s 不能导入数据库。' % (i + 1, field))

    def get_import_data(self, row):
        """
        """
        import_data = {}

        for i in range(len(self.import_index)):
            field = self.import_fields[i]
            index = self.import_index[i]
            val = row[index]
            if not val:
                continue
            if field in self.zhuanhuan:
                # print('转换类型', self.zhuanhuan[field], val)
                val = self.zhuanhuan[field](val)

            import_data[field] = val

        return import_data


def dj_import_csv(ku_model, csv_file):
    """
    导入csv文件到数据库

    注意：需要处理日期和文件名的不要用她导入，她只用于备份和恢复。
    """

    print('导入开始 ---------------------------------------------------------\n', csv_file)

    # 读取csv文件
    with open(csv_file, newline='', encoding='utf-8') as f:
        reader = csv.reader(f)
        try:
            loadList = []
            iter_reader = iter(reader)
            export = ModelData(ku_model, next(iter_reader))
            for row in iter_reader:
                # print(ku_field, xls_row, row)
                kwargs = export.get_import_data(row)
                # print(kwargs)
                loadList.append(ku_model(**kwargs))
        except csv.Error as e:
            return (f'file {csv_file}, line {reader.line_num}: {e}')
        else:
            info = '成功导入 %s 行' % len(ku_model.objects.bulk_create(loadList))
            print(info)
            print('---------------------------------------------------------')
            return info


class LdsExportCsv():
    """
    写入数据到 csv 文件
    支持自定义写入的表头和要导出的字段

    例子
    from ilds.django.model import get_model_verbose_names

    # ku_model = 要导出的模型
    objs = ku_model.objects.all()
    # 找到的歌曲列表
    if objs.exists():
        xlsxfile = r'导出：测试.csv'
        ex_csv = LdsExportCsv(ku_model, xlsxfile)  # 歌曲库的中文名
        ex_csv.set_title_field(*get_model_verbose_names(ku_model))
        # 或者元组列表
        # ex_csv.set_title_field('歌曲名称', '表演者', '专辑名称', ) # , row=3, col=3
        for obj in objs:
            ex_csv.append(obj)
        info = ex_csv.print_count_info()
        ex_csv.close()
    else:
        print('没有找到歌曲')
    """

    def __init__(self, ku_model, work_file):
        """初始化"""
        self.title = []
        self.export = None

        # 要处理的库
        self.ku_model = ku_model

        # 写入csv文件
        self.csvfile = open(work_file, 'w', newline='', encoding='utf-8')  # 有软件打开乱码
        self.writer_csv = csv.writer(self.csvfile)

        self.gequ_dict = {}

        # 设置排除内容
        self.exclude = ['denglushijian_date']  # 'id',

        # 初始化统计
        self.count_all = 0
        self.tongji_info = {}
        self.count_merge = 0

    def set_title_field(self, *export):
        """
        设置标题和写入的内容
        export 是要导出列的名字列表
        """

        # 需要获取数据的字段列表
        self.field_list = []
        self.export = export

        # 获取模型的字段
        model_field = get_model_field(self.ku_model, self.exclude)
        # pprint(model_field)

        # 写表头
        csv_row = []
        for i_export in self.export:
            i_export = i_export.strip()
            if i_export:
                try:
                    self.field_list.append(model_field[i_export])
                except:
                    self.field_list.append('None')
                # print(self.field_list)
                csv_row.append(i_export)
        self.writer_csv.writerow(csv_row)

        self.init_count()

        return

    def append(self, date, row=0, col=0):
        """
        写入 csv 文件
        date 要写入的数据
        """

        model_dict = model_to_dict(date, exclude=self.exclude)

        # 写入文件
        # print(model_dict)
        # print(self.field_list)
        csv_row = []
        for i_field_list in self.field_list:
            try:
                # 统计信息
                if i_field_list in self.tongji_info:
                    value = model_dict[i_field_list]
                    if i_field_list.endswith('_float'):
                        self.tongji_info[i_field_list] += float(value)
                    else:
                        self.tongji_info[i_field_list] += int(value)
                    # print(self.tongji_info)
                # 处理日期
                if i_field_list == 'shujuriqi_date':
                    value = model_dict[i_field_list].strftime("%Y-%m")
                elif i_field_list.endswith('_date'):
                    value = model_dict[i_field_list].strftime("%Y-%m-%d")
                # 处理文件
                elif i_field_list.endswith('_file'):
                    value = get_model_fieldfile_path(model_dict[i_field_list])
                else:
                    value = model_dict[i_field_list]
            except:
                value = ''
            csv_row.append(value)
        self.writer_csv.writerow(csv_row)

        # print(model_dict)
        self.count_all += 1
        return

    def get_count_list(self):
        """获取统计信息的列表 列表内容是字典 键为 count_name count_field count"""
        count_list = []
        model_fields = get_model_field(self.ku_model)
        # print(model_fields)
        if self.tongji_info:
            for k, v in model_fields.items():
                _dict = {}
                _field = self.tongji_info.get(v)
                # print(k,v,_field)
                if not _field is None:
                    _dict['count_name'] = k
                    _dict['count_field'] = v
                    _dict['count'] = self.tongji_info[v]
                    count_list.append(_dict)
        return count_list

    def print_count_info(self):
        """打印统计信息"""
        print_info = '导出：%s    总共：%s' % (self.count_all - self.count_merge, self.count_all)
        for _count in self.get_count_list():
            print_info += '%s：%s    ' % (_count['count_name'], _count['count'])  # _count['count_field']
        print(print_info)
        return print_info

    def init_count(self):
        """ 初始化需要统计的内容"""

        self.tongji_info = {}
        # print(field_list)
        for i in self.field_list:
            if '_quanli_' in i:
                # print('_quanli_')
                continue
            elif i.endswith('_int') or i.endswith('_float'):
                if 'rmb_int' in i:
                    pass
                elif 'fencheng_int' in i:
                    pass
                else:
                    # print(i)
                    self.tongji_info[i] = 0

        # print(self.tongji_info)

    def close(self):
        """
        写入结束，关闭文件
        """
        self.csvfile.close()

    def __repr__(self):
        return str(self.export)


def dj_export_csv(ku_model, sep=''):
    """
    导出数据为csv文件
    """

    name_str = '导出 %s' % ku_model.__name__
    # name_str = '导出 %s(%s)' % (ku_str, ku.__name__)

    EXPOET_EAX = 1000000  #
    data_list = ku_model.objects.all()
    # 自定义了到处方式
    # # 按日期导出
    # kwargs = {}
    # kwargs['shujuriqi_date__range'] = ['2018-01-01', '2100-01-01']
    # data_list = ku.objects.filter(**kwargs)
    try:
        data_count = data_list.count()
    except Exception as e:
        return (ku_model.__name__, '没有歌曲导出 ', e)

    # 找到的歌曲列表
    if data_count:
        print('---------------------------------------------------------')
        print('开始 %s' % name_str)
        print()

        print('搜索结果：', data_count)
        if data_count > EXPOET_EAX:
            export_start = 0
            export_stop = EXPOET_EAX
            info = '已分多个文件导出：'
            # print(info)
            while True:
                if export_start > data_count:
                    break
                data_list = ku_model.objects.all()[export_start:export_stop]
                if export_stop == EXPOET_EAX:
                    name = name_str + '.csv'
                else:
                    # print(type(export_start),type(EXPOET_EAX),type(export_start/EXPOET_EAX))
                    name = name_str + '_%03d.csv' % int(export_start / EXPOET_EAX)
                mycsv = LdsExportCsv(ku_model, name)  # 歌曲库的中文名
                mycsv.set_title_field(*get_model_verbose_names(ku_model, exclude=['id', 'denglushijian_date',
                                                                                  'xiugaizhe']))  # , row=3, col=3
                for i_data_list in data_list:
                    # print(i_data_list)
                    mycsv.append(i_data_list)
                info += sep + mycsv.print_count_info()
                mycsv.close()
                export_start += EXPOET_EAX
                export_stop += EXPOET_EAX
        else:
            print('没有超过 %s' % EXPOET_EAX)
            mycsv = LdsExportCsv(ku_model, name_str + '.csv')  # 歌曲库的中文名
            mycsv.set_title_field(
                *get_model_verbose_names(ku_model, exclude=['id', 'denglushijian_date', 'xiugaizhe']))  # , row=3, col=3
            for i_data_list in data_list:
                # print(i_data_list)
                mycsv.append(i_data_list)
            mycsv.close()
            info = mycsv.print_count_info()
            # print(mycsv.tongji_info)
            # mycsv.print_count_info()
            print()
            print('导出结束')
            print('---------------------------------------------------------')
    else:
        info = (ku_model.__name__, '库没有歌曲')

    # print(info)
    return info


def doc():
    """
    打印模块说明文档
    """
    doc_text = """"""
    doc_text += '\n'
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=ModelData)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=dj_import_csv)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=LdsExportCsv)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=dj_export_csv)
    print(doc_text)


if __name__ == '__main__':
    # 记录运行时间 --------------------------------------------------
    from time import time, sleep

    start_time = t1 = time()

    doc()

    print('运行时间 %.2f 秒' % (time() - start_time))



# -*- coding: utf-8 -*-
#
# ---------------------------------------
#   程序：pd.py
#   版本：0.2
#   作者：lds
#   日期：2019-04-16
#   语言：Python 3.X
#   说明：pandas 常用的函数集合，TODO 添加一些小抄在这里！
# ---------------------------------------
import os

from colorama import Fore, Back, Style

from ilds.file import get_dir_files

try:
    import pandas as pd
except ImportError as e:
    print(Fore.RED + "注：导入 pandas 失败，请安装它: pip install pandas", Style.RESET_ALL)
    pass


def get_columns_index(df, columns):
    """获取列名称在 DataFrame 中的索引"""
    if isinstance(columns, str):
        columns = [columns]
    columns_index = []
    for col in columns:
        columns_index.append(df.columns.get_loc(col))
    # print('columns_index', columns_index)
    return columns_index


def merging_excel_file_data(file_dir, ext='', concat_columns=None, add_source_column=True, is_print=True):
    """
    合并多个 Excel 文件内容

    :param file_dir: 合并文件的路径
    :param ext: 要合并的文件后缀名，默认是文件夹中的全部文件
    :param concat_columns: 指定要合并的列名列表
    :param add_source_column: 是否添加原始来源
    :param is_print: 打印信息
    :return:
    """
    all_len = 0
    frames = []
    for file in get_dir_files(file_dir, ext):
        if is_print:
            print(file)
        with pd.ExcelFile(file) as excel:
            sheet_names = excel.sheet_names
            if is_print:
                print('表薄名称列表：', sheet_names)
            # 通过表薄名字读取表单
            # data['总表'] = pd.read_excel(excel, sheet_name='总表')
            # 读取全部表
            for sheet_name in sheet_names:
                _df = pd.read_excel(excel, sheet_name=sheet_name, header=0)
                # _df = pd.read_excel(excel, sheet_name=0, header=0)
                if is_print:
                    print('表薄名称：', sheet_name, '\t', '行数：', len(_df), '\t', '文件：', file)
                # if 'Metadatas' != sheet_name:
                #     print('跳过内容', sheet_name)
                #     continue
                all_len += len(_df)
                if add_source_column:
                    _df['本行来自'] = f"{os.path.split(file)[1]} - {sheet_name}"
                # 是否指定要合并的列
                if concat_columns is not None:
                    _df = _df[concat_columns]
                frames.append(_df)

    _df = pd.concat(frames)  # result
    if is_print:
        print('合并数据行数：', len(_df), '原始数据行数：', all_len)
    return _df


def writer_excel(obj, path, index=False):
    """
    保存 Excel 文件
    """
    if isinstance(obj, dict):
        with pd.ExcelWriter(path) as writer:
            for sheet_name, df_data in obj.items():
                df_data.to_excel(writer, sheet_name=sheet_name, index=index)
    else:
        with pd.ExcelWriter(path) as writer:
            obj.to_excel(writer, sheet_name='Sheet1', index=index)


def doc():
    """
    打印模块说明文档
    """
    doc_text = """"""
    doc_text += '\n'
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_columns_index)
    # doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=)
    print(doc_text)


if __name__ == '__main__':
    # 记录运行时间 --------------------------------------------------
    from time import time, sleep

    start_time = t1 = time()

    doc()

    print('运行时间 %.2f 秒' % (time() - start_time))



# -*- coding: utf-8 -*-
#
# ---------------------------------------
#   程序：file.py
#   版本：0.7
#   作者：lds
#   日期：2020-03-03
#   语言：Python 3.X
#   说明：常用的文件操作函数集合
# ---------------------------------------


import os
import sys
import random
import shutil
# get_file_md5
import hashlib
# get_encoding
import chardet
# validateTitle
import re
from datetime import datetime
from zlib import crc32
import difflib
import warnings

from colorama import Fore, Back, Style

AFILES = []  # EE
BFILES = []  # SVN
COMMON = []  # EE & SVN


def is_file(file):
    """
    判断是否为文件
    :param file:文件
    :return:文件不存在，不是文件，大小为0，返回 None
    """
    if not os.path.exists(file):
        print(file, ' 没有找到')
        return None
    if not os.path.isfile(file):
        print(file, ' 不是文件')
        return None
    if os.path.getsize(file) == 0:
        print(file, ' 大小为 0')
        return None

    return file


def make_dir(_path):
    """
    检查文件目录是否存在，如果不存在则创建。
    """
    if not os.path.exists(_path):
        os.makedirs(_path)


def exist_or_makedir(in_dir):
    """
    检查文件所在的父级文件夹是否存在，如果不存在，就创建父级文件夹
    :param in_dir:
    :return: None
    """
    warnings.warn('exist_or_makedir 已经弃用，建议用 os.makedirs(name, exist_ok=True) 替代',
                  DeprecationWarning)
    output_dir = os.path.dirname(in_dir)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
        return output_dir


def get_encoding(fromfile):
    """
    文件编码判断
    :param fromfile: 
    :return: 编码格式
    """
    with open(fromfile, 'rb') as tt:
        ff = tt.readline()
        # 这里试着换成read(5)也可以，但是换成readlines()后报错
        enc = chardet.detect(ff)
        # print(enc['encoding'])
        return (enc['encoding'])


def validate_title(title):
    """
    去除文件名中的非法字符 (Windows)
    :param title:
    :return:
    """
    rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
    # 替换为空格，也可以替换为“_”
    new_title = re.sub(rstr, " ", title)
    return new_title


def replace_invalid_filename_char(filename, replaced_char='_', max=100):
    """
    替换文件名中无效的字符。 默认用'_'替换。

    :param filename:  要替换的文件名
    :param replaced_char:  替换的字符
    :return:
    """
    invalid_characaters = '\\/:*?"<>|\r\n\t'
    for c in invalid_characaters:
        filename = filename.replace(c, replaced_char)

    # 清理空白符号
    filename = filename.strip()

    # 截断超过最大长度的字符
    if len(filename) > max:
        filename = f'{filename[:max - 3]}...'

    return filename


def get_file_crc32(file):
    """
    计算文件的 CRC32
    :param file:
    :return:
    """
    with open(file, 'rb') as f:
        return crc32(f.read())


def get_file_md5(file, block_size=4096):
    """
    计算文件的 MD5
    :param file:
    :param block_size:
    :return:
    """

    if not is_file(file):
        return None

    md5_ = hashlib.md5()
    with open(file, 'rb') as f:
        while True:
            data = f.read(block_size)
            if not data:
                break
            md5_.update(data)
    return md5_.hexdigest()


def get_text_md5(text):
    """
    计算文件的 MD5
    :param filename:
    :return:
    """
    return hashlib.md5(text.encode('utf-8')).hexdigest()


def human_size(sz):
    """
    以人类可读的格式返回大小
    """
    if not sz:
        return False
    units = ('bytes', 'Kb', 'Mb', 'Gb')
    s, i = float(sz), 0
    while s >= 1024 and i < len(units) - 1:
        s /= 1024
        i += 1
    return "%0.2f %s" % (s, units[i])


def from_this_dir(file):
    """
    获取运行模块所在路径的全路径
    """
    return os.path.join(os.path.dirname(os.path.dirname(__file__)), file)


def from_this_dir_2(file):
    """
    获取本模块所在路径的全路径
    """
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


def get_file_line_info():
    """ 获取当前时间，文件路径，所在行数"""
    try:
        raise Exception
    except:
        f = sys.exc_info()[2].tb_frame.f_back
    # print(dir(f))
    return '%s, File "%s", line %s ' % (
        str(datetime.now()), f.f_code.co_filename, str(f.f_lineno)
    )
    # , f.f_code.co_name(<module>)


def file_string_replace(file, old, new, count=None):
    """
    按行替换文件内容
    """

    counts = 0
    with open(file, "r", encoding="utf-8") as f:
        # readlines以列表的形式将文件读出
        lines = f.readlines()
    with open(file, "w", encoding="utf-8") as f_w:
        # 定义一个数字，用来记录在读取文件时在列表中的位置
        n = 0
        # 默认选项，只替换第一次匹配到的行中的字符串
        for line in lines:
            if old in line:
                counts += 1
                line = line.replace(old, new)
            f_w.write(line)

    if counts:
        print("在 %s 中替换了 %s 为 %s，总共 %s个" % (file, old, new, counts))
    else:
        print("没有内容替换")


def exists_file(_file):
    """
    判断文件是否存在，如果存在就返回重命名的文件，不存在就直接返回
    :param _file:
    :return:
    """

    warnings.warn('exists_file 已经弃用，因为它对人类不友好，也没有检查新名字是否已经存在\n建议用 check_filename_available 替代',
                  DeprecationWarning)

    if os.path.exists(_file):
        while True:
            f_name, f_ext = os.path.splitext(_file)
            _file2 = f_name + '-' + str(random.randint(0, 10000000)) + f_ext
            if _file != _file2:
                break
        return _file2
    else:
        return _file


def check_filename_available(file):
    """
    检查文件是否存在，如果已经存在，添加编号重命名
    例如: 文件.txt > 文件 (1).txt
    """
    import os
    if not os.path.exists(file):
        return file

    num = 1
    f_name, f_ext = os.path.splitext(file)
    while True:
        new_file_name = f"{f_name} ({num}){f_ext}"
        if not os.path.exists(new_file_name):
            return new_file_name
        num += 1


def exists_file_to_bak(file):
    """
    如果文件已经存在，拷贝到 bak 文件夹
    """
    if os.path.exists(file):
        while True:
            file_path, file_name = os.path.split(file)
            f_name, f_ext = os.path.splitext(file_name)
            _file2 = os.path.join(file_path, 'bak', f_name + '-' + datetime.now().strftime('%Y%m%d%H%M%S') + f_ext)
            if file != _file2:
                break
        make_dir(os.path.dirname(_file2))
        shutil.move(file, _file2)
        # print(shutil.move(_file, _file2))
        # print((_file, _file2))


def get_name(file):
    """
    获取路径中最后的文件名 不包括后缀名
    """
    # os.path.split 分离路径的目录名和文件名(dirname(), basename()) 元组，后一部分总是最后级别的目录或文件名
    _, name = os.path.basename(file)
    # 分离文件名和扩展名， 返回(filename,extension)元组，没有扩展名，扩展名返回空
    ret_name, _ = os.path.splitext(name)
    return ret_name


def list_dir(file_dir):
    """
    获取文件夹下的文件列表
    跳过目录、文件名前缀是.的文件

    :param file_dir:
    :return: 返回文件列表的生成器
    """
    if os.path.isdir(file_dir):
        for name in os.listdir(file_dir):
            file = os.path.join(file_dir, name)

            # 跳过目录
            if os.path.isdir(file):
                continue

            # 跳过文件名前缀是.的文件
            if name.startswith('.'):
                continue

            # print(file)
            yield file
    else:
        raise NotADirectoryError(file_dir)


def from_dir_func(dir_path, func, prefix='.', suffix='', *args, **kwargs):
    """

    处理目录中的所有文件，默认跳过前缀是.的文件名，返回函数运行结果

    # 扩展阅读
    from functools import partial
    基于一个函数创建一个新的可调用对象，把原函数的某些参数固定。
    使用这个函数可以把接受一个或多个参数的函数改编成需要回调的 API，这样参数更少。
    new_func = partial(test, 22222)

    :param dir_path:
    :param func:
    :param prefix: 默认跳过前缀是.的文件名
    :param suffix: 只处理后缀匹配的文件
    :param args:
    :param kwargs:
    :return: 返回函数运行结果的信息
    """

    fileok = 0
    fileno = 0
    info = []
    if os.path.isdir(dir_path):
        print('\n处理路径：\n%s\n' % (dir_path))
        for dirpath, dirnames, filenames in os.walk(dir_path):
            for _filename in filenames:
                if prefix and _filename.startswith(prefix):
                    continue

                _file = os.path.join(dirpath, _filename)
                if suffix:
                    if _file.endswith(suffix):
                        fileok += 1
                        # print(_file)
                        info.append(func(_file, *args, **kwargs))
                    else:
                        fileno += 1
                        # print('忽略 --------------------', _file)
                else:
                    fileok += 1
                    info.append(func(_file, *args, **kwargs))

        print(' ----------- 处理 %s 个文件（跳过名称前面是：“%s”，处理后缀：“%s”） ----------- 忽略 %s 个文件 ----------- ' % (
            fileok, prefix, suffix, fileno))
    else:
        raise FileExistsError('请输入文件路径！')
    return info


def get_walk_files(dir_path, endswith=''):
    """
    获取文件夹包括子文件夹里面的文件列表（生成器）
    """
    if os.path.isdir(dir_path):
        print('\n处理路径：\n%s\n' % (dir_path))
        for dirpath, dirnames, filenames in os.walk(dir_path):
            for _filename in filenames:
                if _filename.startswith('.'):
                    continue

                _file = os.path.join(dirpath, _filename)
                if endswith:
                    if _file.endswith(endswith):
                        yield _file
                else:
                    yield _file
    else:
        return None


def get_dir_files(path, ext=''):
    """
    获取文件夹里面，指定后缀名的文件列表（生成器）
    例子：
    get_dir_files(r".",'py')
    """
    # files = []
    for _file in os.listdir(path):
        if _file.endswith(ext):
            file_path = os.path.join(path, _file)
            if os.path.isfile(file_path):
                # files.append(file_path)
                yield file_path
                # print(_file)
    # return files


def save_file(s, file, mode='w', encoding='utf-8'):
    """保存字符内容到文件"""
    with open(file, mode, encoding=encoding) as fp:
        fp.write(s)


def dir_compare(apath, bpath, diff_ext=None, out_dir=None):
    """
    比较两个目录的文件差异

    例子：
    diff_ext = ['.md']
    dir_compare(FolderEE, FolderSVN, diff_ext)

    :param apath:
    :param bpath:
    :param diff_ext:
    :return:
    """

    if diff_ext is None:
        diff_ext = []

    if out_dir is None:
        out_dir = os.getcwd()

    afiles = []
    bfiles = []
    for root, dirs, files in os.walk(apath):
        # print(apath, '所有文件数量：', len(files))
        for f in files:
            # 比较文件名不含格式后缀
            # afiles.append(root + f[0:-4])
            # 比较文件名含格式后缀
            afiles.append(os.path.join(root, f))

    for root, dirs, files in os.walk(bpath):
        # print(bpath, '所有文件数量：', len(files))
        for f in files:
            # 比较文件名不含格式后缀
            # bfiles.append(root + f[0:-4])

            # 比较文件名含格式后缀
            bfiles.append(os.path.join(root, f))
            # sizeB = os.path.getsize(root + "/" + f) 此处定义的size无法在commonfiles进行比较. (A,B在各自的循环里面)

    # print(afiles, bfiles)

    # 去掉 afiles 中文件名的 apath (拿A,B相同的路径\文件名,做成集合,去找交集)
    apathlen = len(apath)
    aafiles = []
    for f in afiles:
        aafiles.append(f[apathlen:])

    # 去掉 bfiles 中文件名的  bpath
    bpathlen = len(bpath)
    bbfiles = []
    for f in bfiles:
        bbfiles.append(f[bpathlen:])

    afiles = aafiles
    bfiles = bbfiles

    setA = set(afiles)
    setB = set(bfiles)
    # print('%$%'+str(len(setA)))
    # print('%%'+str(len(setB)))
    commonfiles = setA & setB  # 处理共有文件
    # print ("===============File with different size in '", apath, "' and '", bpath, "'===============")
    # 将结果输出到本地
    # with open(os.getcwd()+'diff.txt','w') as di:
    # di.write("===============File with different size in '", apath, "' and '", bpath, "'===============")
    # print(commonfiles)

    diff_info = []
    for f in sorted(commonfiles):
        a_file = os.path.join(apath + f)
        b_file = os.path.join(bpath + f)
        # print(apath, f, a_file, b_file)

        a_file_size = os.path.getsize(a_file)
        b_file_size = os.path.getsize(b_file)

        # return
        if a_file_size == b_file_size:  # 共有文件的大小比较
            # pass #print (f + "\t\t" + get_pretty_time(os.stat(a_file)) + "\t\t" + get_pretty_time(os.stat(b_file)))
            # 以下代码是处理大小一致，但是内容可能不一致的情况，比较 md5 需要的时间比较长
            a_file_md5 = get_file_md5(a_file)
            b_file_md5 = get_file_md5(b_file)
            if a_file_md5 == b_file_md5:
                continue
            else:
                # Git用<<<<<<<，=======，>>>>>>>标记出不同分支的内容。HEAD为当前所在分支的内容，也就是说现在master中的内容
                info = f'{"=" * 70}  文件MD5: {a_file_md5} != {b_file_md5}\n{f}\n\n'
                # "文件名=%s    MD5不同，文件 A:%s   !=  文件 B:%s" % (f, a_file_md5, b_file_md5)
            # print(os.getcwd())

        else:
            info = f'{"=" * 70}  文件大小: {a_file_size} != {b_file_size}\n{f}\n\n'

        # 只处理指定后缀的内容差异
        if os.path.splitext(a_file)[1].lower() in diff_ext:
            with open(a_file, 'r', encoding='utf-8') as f_in:
                AText = f_in.read()
            with open(b_file, 'r', encoding='utf-8') as f_in:
                BText = f_in.read()
            differ = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
            diff = differ.compare(
                AText.splitlines(keepends=True), BText.splitlines(keepends=True)  # keepends 包含换行符
            )
            diff = [d for d in diff if d.startswith('+') or d.startswith('-')]
            # print(''.join(diff))
            info += ''.join(diff)
            # if len(list(diff)) < 100:
            #     print(''.join(diff))
            #     exit()

        # 文件不同的时候处理
        diff_info.append(info)
        print(info)

    diff_file = os.path.join(out_dir, 'diff.txt')
    if os.path.exists(diff_file):
        os.remove(diff_file)
    if diff_info:
        with open(diff_file, 'a', encoding='utf-8') as di:
            di.write('\n'.join(diff_info))
    else:
        if diff_ext:
            print(f'文件夹中的 {diff_ext} 后缀名文件没有差异')

    # 处理仅出现在一个目录中的文件
    onlyFiles = setA ^ setB
    aonlyFiles = []
    bonlyFiles = []
    for of in onlyFiles:
        if of in afiles:
            aonlyFiles.append(of)
        elif of in bfiles:
            bonlyFiles.append(of)

    # print ("###################### EE resource ONLY ###########################")
    # print ("#only files in ", apath)
    a_only_file = os.path.join(out_dir, os.path.basename(apath) + ' only.txt')
    b_only_file = os.path.join(out_dir, os.path.basename(bpath) + ' only.txt')
    if os.path.exists(a_only_file):
        os.remove(a_only_file)
    if os.path.exists(b_only_file):
        os.remove(b_only_file)

    aonly_count = len(aonlyFiles)
    bonly_count = len(bonlyFiles)

    if aonly_count:
        with open(a_only_file, 'a') as a:
            for of in sorted(aonlyFiles):
                a.write(of + '\n')
                # a.write(apath + of + '\n')

        # print (of)
    # print ("*"*20+"SVN ONLY+"+"*"*20)
    # print ("#only files in ", bpath)

    if bonly_count:
        with open(b_only_file, 'a') as b:
            for of in sorted(bonlyFiles):
                b.write(of + '\n')
                # b.write(bpath + of + '\n')
            # print (of)

    print(apath, 'only files numbers:', aonly_count)
    print(bpath, 'only files numbers:', bonly_count)


def get_compound_file_binary(file):
    """
    获取复合文件二进制格式文件中的数据
    
    Compound File Binary Format Files
    https://stackoverflow.com/questions/12705527/reading-excel-files-with-xlrd
    """
    try:
        import olefile
        with open(file, 'rb') as f:
            ole = olefile.OleFileIO(f)
            # print(ole.listdir())
            if ole.exists('Workbook'):
                d = ole.openstream('Workbook')
                return d
            else:
                return f
    except ImportError as e:
        print(Fore.RED + "注：找不到 olefile，请安装它: pip install olefile", Style.RESET_ALL)
        pass


def doc():
    """
    打印模块说明文档
    """
    doc_text = """
    # 默认字典的例子
    from collections import defaultdict
    def get_counts2(sequence):
        counts = defaultdict(int) # 所有的值均会被初始化为0
        for x in sequence:
            counts[x] += 1
        return counts
    
    """
    doc_text += '\n'
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=is_file)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=make_dir)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=exist_or_makedir)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_encoding)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=validate_title)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=replace_invalid_filename_char)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_file_crc32)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_file_md5)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_text_md5)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=human_size)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=from_this_dir)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=from_this_dir_2)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_file_line_info)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=file_string_replace)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=exists_file)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=exists_file_to_bak)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=get_name)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=list_dir)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=from_dir_func)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=save_file)
    doc_text += '{fun.__name__}{fun.__doc__}\n'.format(fun=dir_compare)
    print(doc_text)


if __name__ == '__main__':
    # 记录运行时间 --------------------------------------------------
    from time import time, sleep

    start_time = t1 = time()
    get_compound_file_binary('compound_file')
    doc()

    # 性能测试
    # from timeit import timeit
    # print(timeit('validate_title("\\fasdf/f:*?dasfdddddddddddddddddddddda<fasf>|")', 'from file import validate_title', number=1000000))
    # replace_invalid_filename_char 稍微快一点
    # print(timeit('replace_invalid_filename_char("\\fasdf/f:*?dasfdddddddddddddddddddddda<fasf>|")', 'from file import replace_invalid_filename_char', number=1000000))

    print('运行时间 %.2f 秒' % (time() - start_time))



# -*- coding: utf-8 -*-
#
# ---------------------------------------
#   程序：mycsv.py
#   版本：0.1
#   作者：lds
#   日期：2019-10-11
#   语言：Python 3.X
#   说明：处理 csv 的函数集合
# ---------------------------------------

import os
import csv

import xlwt
import xlsxwriter

from ilds.file import get_encoding


def csv_to_xls(csv_file, delimiter=None):
    """
    cvs 转换为 xls 文件
    """

    # 获取需要导出的文件名
    file_path, file_name = os.path.split(csv_file)
    file_name = os.path.splitext(file_name)[0] + '.xls'
    file_path = os.path.join(file_path, 'xls')
    if not os.path.exists(file_path):
        os.makedirs(file_path)

    xls_file = os.path.join(file_path, file_name)

    print(csv_file, 'to', xls_file)

    workbook = xlwt.Workbook(encoding='utf-8')
    sheet = workbook.add_sheet('csv to xls', cell_overwrite_ok=True)

    # with open(csvfile, newline='') as f: # ,encoding='utf-8'
    # print(get_encoding(csvfile))
    encoding = get_encoding(csv_file)
    if encoding.lower() == 'gb2312':
        encoding = 'gbk'
    f = open(csv_file, newline='', encoding=encoding)
    if delimiter is None:
        reader = csv.reader(f)
    else:
        reader = csv.reader(f, delimiter=delimiter)
    xls_sheet = 1
    line = 0
    xls_tite = None
    for row in reader:
        if line == 0:
            if xls_tite is None:
                xls_tite = row
                # print(xls_tite)
        elif line >= 65535:
            line = 0
            xls_sheet += 1
            sheet = workbook.add_sheet('cvs to xls ' + str(xls_sheet), cell_overwrite_ok=True)
            for c, col in enumerate(xls_tite):
                sheet.write(line, c, col.strip())
            line += 1
        for c, col in enumerate(row):
            sheet.write(line, c, col.strip())
        line += 1

    # 关闭文件
    workbook.save(xls_file)
    f.close()

    return xls_file


def csv_to_xlsx(csv_file, delimiter=None):
    """
    cvs 转换为 xlsx 文件
    """

    # 获取需要导出的文件名
    file_path, file_name = os.path.split(csv_file)
    file_name = os.path.splitext(file_name)[0] + '.xlsx'
    file_path = os.path.join(file_path, 'xlsx')
    if not os.path.exists(file_path):
        os.makedirs(file_path)

    xlsx_file = os.path.join(file_path, file_name)

    print(csv_file, 'to', xlsx_file)

    workbook = xlsxwriter.Workbook(xlsx_file)
    worksheet = workbook.add_worksheet()
    worksheet.name = 'cvs to xlsx'

    # with open(csvfile, newline='') as f: # ,encoding='utf-8'
    encoding = get_encoding(csv_file)
    if encoding.lower() == 'gb2312':
        encoding = 'gbk'
    f = open(csv_file, newline='', encoding=encoding)
    if delimiter is None:
        reader = csv.reader(f)
    else:
        reader = csv.reader(f, delimiter=delimiter)

    line = 0
    for row in reader:
        # print(row, len(row))
        for i, cell in enumerate(row):
            # print(cell)
            cell = cell.strip()
            # print(cell)
            worksheet.write(line, i, cell)
        line += 1
    workbook.close()
    f.close()

    return xlsx_file


if __name__ == '__main__':
    from ilds.time import Timer

    with Timer() as timer:
        ...
        csv_to_xls(r"D:\2019-10-11 16-34-44.csv", delimiter=None)
        csv_to_xlsx(r"D:\2019-10-11 16-34-44.csv", delimiter=None)
